@@ -55,7 +55,7 @@
  * This code isn't concerned about the FSM at all. The caller is responsible
  * for initializing that.
  *
- * Portions Copyright (c) 1996-2014, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2015, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -66,12 +66,14 @@
 
 #include "postgres.h"
 
-#include "access/heapam_xlog.h"
 #include "access/nbtree.h"
+#include "access/xlog.h"
+#include "access/xloginsert.h"
 #include "miscadmin.h"
 #include "storage/smgr.h"
 #include "tcop/tcopprot.h"
 #include "utils/rel.h"
+#include "utils/sortsupport.h"
 #include "utils/tuplesort.h"
 
 
@@ -185,9 +187,10 @@ _bt_spooldestroy(BTSpool *btspool)
  * spool an index entry into the sort file.
  */
 void
-_bt_spool(IndexTuple itup, BTSpool *btspool)
+_bt_spool(BTSpool *btspool, ItemPointer self, Datum *values, bool *isnull)
 {
-	tuplesort_putindextuple(btspool->sortstate, itup);
+	tuplesort_putindextuplevalues(btspool->sortstate, btspool->index,
+								  self, values, isnull);
 }
 
 /*
@@ -684,6 +687,7 @@ _bt_load(BTWriteState *wstate, BTSpool *btspool, BTSpool *btspool2)
 	int			i,
 				keysz = RelationGetNumberOfAttributes(wstate->index);
 	ScanKey		indexScanKey = NULL;
+	SortSupport sortKeys;
 
 	if (merge)
 	{
@@ -699,6 +703,33 @@ _bt_load(BTWriteState *wstate, BTSpool *btspool, BTSpool *btspool2)
 										true, &should_free2);
 		indexScanKey = _bt_mkscankey_nodata(wstate->index);
 
+		/* Prepare SortSupport data for each column */
+		sortKeys = (SortSupport) palloc0(keysz * sizeof(SortSupportData));
+
+		for (i = 0; i < keysz; i++)
+		{
+			SortSupport sortKey = sortKeys + i;
+			ScanKey		scanKey = indexScanKey + i;
+			int16		strategy;
+
+			sortKey->ssup_cxt = CurrentMemoryContext;
+			sortKey->ssup_collation = scanKey->sk_collation;
+			sortKey->ssup_nulls_first =
+				(scanKey->sk_flags & SK_BT_NULLS_FIRST) != 0;
+			sortKey->ssup_attno = scanKey->sk_attno;
+			/* Abbreviation is not supported here */
+			sortKey->abbreviate = false;
+
+			AssertState(sortKey->ssup_attno != 0);
+
+			strategy = (scanKey->sk_flags & SK_BT_DESC) != 0 ?
+				BTGreaterStrategyNumber : BTLessStrategyNumber;
+
+			PrepareSortSupportFromIndexRel(wstate->index, strategy, sortKey);
+		}
+
+		_bt_freeskey(indexScanKey);
+
 		for (;;)
 		{
 			load1 = true;		/* load BTSpool next ? */
@@ -711,43 +742,20 @@ _bt_load(BTWriteState *wstate, BTSpool *btspool, BTSpool *btspool2)
 			{
 				for (i = 1; i <= keysz; i++)
 				{
-					ScanKey		entry;
+					SortSupport entry;
 					Datum		attrDatum1,
 								attrDatum2;
 					bool		isNull1,
 								isNull2;
 					int32		compare;
 
-					entry = indexScanKey + i - 1;
+					entry = sortKeys + i - 1;
 					attrDatum1 = index_getattr(itup, i, tupdes, &isNull1);
 					attrDatum2 = index_getattr(itup2, i, tupdes, &isNull2);
-					if (isNull1)
-					{
-						if (isNull2)
-							compare = 0;		/* NULL "=" NULL */
-						else if (entry->sk_flags & SK_BT_NULLS_FIRST)
-							compare = -1;		/* NULL "<" NOT_NULL */
-						else
-							compare = 1;		/* NULL ">" NOT_NULL */
-					}
-					else if (isNull2)
-					{
-						if (entry->sk_flags & SK_BT_NULLS_FIRST)
-							compare = 1;		/* NOT_NULL ">" NULL */
-						else
-							compare = -1;		/* NOT_NULL "<" NULL */
-					}
-					else
-					{
-						compare =
-							DatumGetInt32(FunctionCall2Coll(&entry->sk_func,
-														 entry->sk_collation,
-															attrDatum1,
-															attrDatum2));
 
-						if (entry->sk_flags & SK_BT_DESC)
-							INVERT_COMPARE_RESULT(compare);
-					}
+					compare = ApplySortComparator(attrDatum1, isNull1,
+												  attrDatum2, isNull2,
+												  entry);
 					if (compare > 0)
 					{
 						load1 = false;
@@ -781,7 +789,7 @@ _bt_load(BTWriteState *wstate, BTSpool *btspool, BTSpool *btspool2)
 												true, &should_free2);
 			}
 		}
-		_bt_freeskey(indexScanKey);
+		pfree(sortKeys);
 	}
 	else
 	{
