@@ -14,11 +14,15 @@
 #
 #----------------------------------------------------------------------
 
-use Catalog;
-
 use strict;
 use warnings;
 
+use File::Basename;
+use File::Spec;
+BEGIN  { use lib File::Spec->rel2abs(dirname(__FILE__)); }
+
+use Catalog;
+
 my @input_files;
 our @include_path;
 my $output_path = '';
