@@ -355,6 +355,12 @@ ExecScanReScan(ScanState *node)
 	/* Stop projecting any tuples from SRFs in the targetlist */
 	node->ps.ps_TupFromTlist = false;
 
+	/*
+	 * We must clear the scan tuple so that observers (e.g., execCurrent.c)
+	 * can tell that this plan node is not positioned on a tuple.
+	 */
+	ExecClearTuple(node->ss_ScanTupleSlot);
+
 	/* Rescan EvalPlanQual tuple if we're inside an EvalPlanQual recheck */
 	if (estate->es_epqScanDone != NULL)
 	{
