@@ -23,18 +23,29 @@ static void tsearch_readline_callback(void *arg);
 
 #ifdef USE_WIDE_UPPER_LOWER
 
+/*
+ * The reason these functions use a 3-wchar_t output buffer, not 2 as you
+ * might expect, is that on Windows "wchar_t" is 16 bits and what we'll be
+ * getting from char2wchar() is UTF16 not UTF32.  A single input character
+ * may therefore produce a surrogate pair rather than just one wchar_t;
+ * we also need room for a trailing null.  When we do get a surrogate pair,
+ * we pass just the first code to iswdigit() etc, so that these functions will
+ * always return false for characters outside the Basic Multilingual Plane.
+ */
+#define WC_BUF_LEN  3
+
 int
 t_isdigit(const char *ptr)
 {
 	int			clen = pg_mblen(ptr);
-	wchar_t		character[2];
+	wchar_t		character[WC_BUF_LEN];
 	Oid			collation = DEFAULT_COLLATION_OID;		/* TODO */
 	pg_locale_t mylocale = 0;	/* TODO */
 
 	if (clen == 1 || lc_ctype_is_c(collation))
 		return isdigit(TOUCHAR(ptr));
 
-	char2wchar(character, 2, ptr, clen, mylocale);
+	char2wchar(character, WC_BUF_LEN, ptr, clen, mylocale);
 
 	return iswdigit((wint_t) character[0]);
 }
@@ -43,14 +54,14 @@ int
 t_isspace(const char *ptr)
 {
 	int			clen = pg_mblen(ptr);
-	wchar_t		character[2];
+	wchar_t		character[WC_BUF_LEN];
 	Oid			collation = DEFAULT_COLLATION_OID;		/* TODO */
 	pg_locale_t mylocale = 0;	/* TODO */
 
 	if (clen == 1 || lc_ctype_is_c(collation))
 		return isspace(TOUCHAR(ptr));
 
-	char2wchar(character, 2, ptr, clen, mylocale);
+	char2wchar(character, WC_BUF_LEN, ptr, clen, mylocale);
 
 	return iswspace((wint_t) character[0]);
 }
@@ -59,14 +70,14 @@ int
 t_isalpha(const char *ptr)
 {
 	int			clen = pg_mblen(ptr);
-	wchar_t		character[2];
+	wchar_t		character[WC_BUF_LEN];
 	Oid			collation = DEFAULT_COLLATION_OID;		/* TODO */
 	pg_locale_t mylocale = 0;	/* TODO */
 
 	if (clen == 1 || lc_ctype_is_c(collation))
 		return isalpha(TOUCHAR(ptr));
 
-	char2wchar(character, 2, ptr, clen, mylocale);
+	char2wchar(character, WC_BUF_LEN, ptr, clen, mylocale);
 
 	return iswalpha((wint_t) character[0]);
 }
@@ -75,14 +86,14 @@ int
 t_isprint(const char *ptr)
 {
 	int			clen = pg_mblen(ptr);
-	wchar_t		character[2];
+	wchar_t		character[WC_BUF_LEN];
 	Oid			collation = DEFAULT_COLLATION_OID;		/* TODO */
 	pg_locale_t mylocale = 0;	/* TODO */
 
 	if (clen == 1 || lc_ctype_is_c(collation))
 		return isprint(TOUCHAR(ptr));
 
-	char2wchar(character, 2, ptr, clen, mylocale);
+	char2wchar(character, WC_BUF_LEN, ptr, clen, mylocale);
 
 	return iswprint((wint_t) character[0]);
 }
@@ -139,10 +150,28 @@ tsearch_readline(tsearch_readline_state *stp)
 {
 	char	   *result;
 
+	/* Advance line number to use in error reports */
 	stp->lineno++;
-	stp->curline = NULL;
+
+	/* Clear curline, it's no longer relevant */
+	if (stp->curline)
+	{
+		pfree(stp->curline);
+		stp->curline = NULL;
+	}
+
+	/* Collect next line, if there is one */
 	result = t_readline(stp->fp);
-	stp->curline = result;
+	if (!result)
+		return NULL;
+
+	/*
+	 * Save a copy of the line for possible use in error reports.  (We cannot
+	 * just save "result", since it's likely to get pfree'd at some point by
+	 * the caller; an error after that would try to access freed data.)
+	 */
+	stp->curline = pstrdup(result);
+
 	return result;
 }
 
@@ -152,7 +181,16 @@ tsearch_readline(tsearch_readline_state *stp)
 void
 tsearch_readline_end(tsearch_readline_state *stp)
 {
+	/* Suppress use of curline in any error reported below */
+	if (stp->curline)
+	{
+		pfree(stp->curline);
+		stp->curline = NULL;
+	}
+
+	/* Release other resources */
 	FreeFile(stp->fp);
+
 	/* Pop the error context stack */
 	error_context_stack = stp->cb.previous;
 }
