@@ -2,7 +2,7 @@
  *
  *	  EUC_TW <--> UTF8
  *
- * Portions Copyright (c) 1996-2014, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2015, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -22,9 +22,6 @@ PG_MODULE_MAGIC;
 PG_FUNCTION_INFO_V1(euc_tw_to_utf8);
 PG_FUNCTION_INFO_V1(utf8_to_euc_tw);
 
-extern Datum euc_tw_to_utf8(PG_FUNCTION_ARGS);
-extern Datum utf8_to_euc_tw(PG_FUNCTION_ARGS);
-
 /* ----------
  * conv_proc(
  *		INTEGER,	-- source encoding id
@@ -44,8 +41,11 @@ euc_tw_to_utf8(PG_FUNCTION_ARGS)
 
 	CHECK_ENCODING_CONVERSION_ARGS(PG_EUC_TW, PG_UTF8);
 
-	LocalToUtf(src, dest, LUmapEUC_TW, NULL,
-		   sizeof(LUmapEUC_TW) / sizeof(pg_local_to_utf), 0, PG_EUC_TW, len);
+	LocalToUtf(src, len, dest,
+			   LUmapEUC_TW, lengthof(LUmapEUC_TW),
+			   NULL, 0,
+			   NULL,
+			   PG_EUC_TW);
 
 	PG_RETURN_VOID();
 }
@@ -59,8 +59,11 @@ utf8_to_euc_tw(PG_FUNCTION_ARGS)
 
 	CHECK_ENCODING_CONVERSION_ARGS(PG_UTF8, PG_EUC_TW);
 
-	UtfToLocal(src, dest, ULmapEUC_TW, NULL,
-		   sizeof(ULmapEUC_TW) / sizeof(pg_utf_to_local), 0, PG_EUC_TW, len);
+	UtfToLocal(src, len, dest,
+			   ULmapEUC_TW, lengthof(ULmapEUC_TW),
+			   NULL, 0,
+			   NULL,
+			   PG_EUC_TW);
 
 	PG_RETURN_VOID();
 }
