@@ -2,7 +2,7 @@
  *
  *	  SHIFT_JIS_2004 <--> UTF8
  *
- * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2015, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -24,9 +24,6 @@ PG_MODULE_MAGIC;
 PG_FUNCTION_INFO_V1(shift_jis_2004_to_utf8);
 PG_FUNCTION_INFO_V1(utf8_to_shift_jis_2004);
 
-extern Datum shift_jis_2004_to_utf8(PG_FUNCTION_ARGS);
-extern Datum utf8_to_shift_jis_2004(PG_FUNCTION_ARGS);
-
 /* ----------
  * conv_proc(
  *		INTEGER,	-- source encoding id
@@ -46,10 +43,11 @@ shift_jis_2004_to_utf8(PG_FUNCTION_ARGS)
 
 	CHECK_ENCODING_CONVERSION_ARGS(PG_SHIFT_JIS_2004, PG_UTF8);
 
-	LocalToUtf(src, dest, LUmapSHIFT_JIS_2004, LUmapSHIFT_JIS_2004_combined,
-			   sizeof(LUmapSHIFT_JIS_2004) / sizeof(pg_local_to_utf),
-	 sizeof(LUmapSHIFT_JIS_2004_combined) / sizeof(pg_local_to_utf_combined),
-			   PG_SHIFT_JIS_2004, len);
+	LocalToUtf(src, len, dest,
+			   LUmapSHIFT_JIS_2004, lengthof(LUmapSHIFT_JIS_2004),
+		LUmapSHIFT_JIS_2004_combined, lengthof(LUmapSHIFT_JIS_2004_combined),
+			   NULL,
+			   PG_SHIFT_JIS_2004);
 
 	PG_RETURN_VOID();
 }
@@ -63,10 +61,11 @@ utf8_to_shift_jis_2004(PG_FUNCTION_ARGS)
 
 	CHECK_ENCODING_CONVERSION_ARGS(PG_UTF8, PG_SHIFT_JIS_2004);
 
-	UtfToLocal(src, dest, ULmapSHIFT_JIS_2004, ULmapSHIFT_JIS_2004_combined,
-			   sizeof(ULmapSHIFT_JIS_2004) / sizeof(pg_utf_to_local),
-	 sizeof(ULmapSHIFT_JIS_2004_combined) / sizeof(pg_utf_to_local_combined),
-			   PG_SHIFT_JIS_2004, len);
+	UtfToLocal(src, len, dest,
+			   ULmapSHIFT_JIS_2004, lengthof(ULmapSHIFT_JIS_2004),
+		ULmapSHIFT_JIS_2004_combined, lengthof(ULmapSHIFT_JIS_2004_combined),
+			   NULL,
+			   PG_SHIFT_JIS_2004);
 
 	PG_RETURN_VOID();
 }
