@@ -201,6 +201,11 @@ tablespace_list_append(const char *arg)
 		exit(1);
 	}
 
+	/*
+	 * Comparisons done with these values should involve similarly
+	 * canonicalized path values.  This is particularly sensitive on Windows
+	 * where path values may not necessarily use Unix slashes.
+	 */
 	canonicalize_path(cell->old_dir);
 	canonicalize_path(cell->new_dir);
 
@@ -735,8 +740,12 @@ writeTarData(
 #ifdef HAVE_LIBZ
 	if (ztarfile != NULL)
 	{
+		errno = 0;
 		if (gzwrite(ztarfile, buf, r) != r)
 		{
+			/* if write didn't set errno, assume problem is no disk space */
+			if (errno == 0)
+				errno = ENOSPC;
 			fprintf(stderr,
 					_("%s: could not write to compressed file \"%s\": %s\n"),
 					progname, current_file, get_gz_error(ztarfile));
@@ -746,8 +755,12 @@ writeTarData(
 	else
 #endif
 	{
+		errno = 0;
 		if (fwrite(buf, r, 1, tarfile) != 1)
 		{
+			/* if write didn't set errno, assume problem is no disk space */
+			if (errno == 0)
+				errno = ENOSPC;
 			fprintf(stderr, _("%s: could not write to file \"%s\": %s\n"),
 					progname, current_file, strerror(errno));
 			disconnect_and_exit(1);
@@ -1120,9 +1133,14 @@ static const char *
 get_tablespace_mapping(const char *dir)
 {
 	TablespaceListCell *cell;
+	char		canon_dir[MAXPGPATH];
+
+	/* Canonicalize path for comparison consistency */
+	strlcpy(canon_dir, dir, sizeof(canon_dir));
+	canonicalize_path(canon_dir);
 
 	for (cell = tablespace_dirs.head; cell; cell = cell->next)
-		if (strcmp(dir, cell->old_dir) == 0)
+		if (strcmp(canon_dir, cell->old_dir) == 0)
 			return cell->new_dir;
 
 	return dir;
@@ -1347,8 +1365,12 @@ ReceiveAndUnpackTarFile(PGconn *conn, PGresult *res, int rownum)
 				continue;
 			}
 
+			errno = 0;
 			if (fwrite(copybuf, r, 1, file) != 1)
 			{
+				/* if write didn't set errno, assume problem is no disk space */
+				if (errno == 0)
+					errno = ENOSPC;
 				fprintf(stderr, _("%s: could not write to file \"%s\": %s\n"),
 						progname, filename, strerror(errno));
 				disconnect_and_exit(1);
@@ -1826,7 +1848,7 @@ BaseBackup(void)
 	{
 #ifndef WIN32
 		int			status;
-		int			r;
+		pid_t		r;
 #else
 		DWORD		status;
 		/*
@@ -1853,7 +1875,7 @@ BaseBackup(void)
 
 		/* Just wait for the background process to exit */
 		r = waitpid(bgchild, &status, 0);
-		if (r == -1)
+		if (r == (pid_t) -1)
 		{
 			fprintf(stderr, _("%s: could not wait for child process: %s\n"),
 					progname, strerror(errno));
@@ -1862,19 +1884,13 @@ BaseBackup(void)
 		if (r != bgchild)
 		{
 			fprintf(stderr, _("%s: child %d died, expected %d\n"),
-					progname, r, (int) bgchild);
+					progname, (int) r, (int) bgchild);
 			disconnect_and_exit(1);
 		}
-		if (!WIFEXITED(status))
+		if (status != 0)
 		{
-			fprintf(stderr, _("%s: child process did not exit normally\n"),
-					progname);
-			disconnect_and_exit(1);
-		}
-		if (WEXITSTATUS(status) != 0)
-		{
-			fprintf(stderr, _("%s: child process exited with error %d\n"),
-					progname, WEXITSTATUS(status));
+			fprintf(stderr, "%s: %s\n",
+					progname, wait_result_to_str(status));
 			disconnect_and_exit(1);
 		}
 		/* Exited normally, we're happy! */
@@ -2237,7 +2253,7 @@ main(int argc, char **argv)
 			exit(1);
 		}
 #else
-		fprintf(stderr, _("%s: symlinks are not supported on this platform\n"));
+		fprintf(stderr, _("%s: symlinks are not supported on this platform\n"), progname);
 		exit(1);
 #endif
 		free(linkloc);
