@@ -215,7 +215,7 @@ main(int argc, char *argv[])
 		values[2] = options.username;
 		keywords[3] = "password";
 		values[3] = password;
-		keywords[4] = "dbname";
+		keywords[4] = "dbname"; /* see do_connect() */
 		values[4] = (options.action == ACT_LIST_DB &&
 					 options.dbname == NULL) ?
 			"postgres" : options.dbname;
@@ -571,15 +571,18 @@ parse_psql_options(int argc, char *argv[], struct adhoc_opts * options)
 				options->single_txn = true;
 				break;
 			case '?':
-				/* Actual help option given */
-				if (strcmp(argv[optind - 1], "-?") == 0)
+				if (optind <= argc &&
+					strcmp(argv[optind - 1], "-?") == 0)
 				{
+					/* actual help option given */
 					usage(NOPAGER);
 					exit(EXIT_SUCCESS);
 				}
-				/* unknown option reported by getopt */
 				else
+				{
+					/* getopt error (unknown option or missing argument) */
 					goto unknown_option;
+				}
 				break;
 			case 1:
 				{
