@@ -264,6 +264,24 @@ SetVariableAssignHook(VariableSpace space, const char *name, VariableAssignHook
 	return true;
 }
 
+/*
+ * Return true iff the named variable has an assign hook function.
+ */
+bool
+VariableHasHook(VariableSpace space, const char *name)
+{
+	struct _variable *current;
+
+	Assert(space);
+	Assert(name);
+
+	for (current = space->next; current; current = current->next)
+		if (strcmp(current->name, name) == 0)
+			return current->assign_hook != NULL;
+
+	return false;
+}
+
 bool
 SetVariableBool(VariableSpace space, const char *name)
 {
