@@ -70,7 +70,7 @@ extern size_t pg_strftime(char *s, size_t max, const char *format,
 
 /* these functions and variables are in pgtz.c */
 
-extern pg_tz *session_timezone;
+extern PGDLLIMPORT pg_tz *session_timezone;
 extern pg_tz *log_timezone;
 
 extern void pg_timezone_initialize(void);
