@@ -28,260 +28,182 @@ ecpg_raise(int line, int code, const char *sqlstate, const char *str)
 	{
 		case ECPG_NOT_FOUND:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("no data found on line %d"), line);
 			break;
 
 		case ECPG_OUT_OF_MEMORY:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("out of memory on line %d"), line);
 			break;
 
 		case ECPG_UNSUPPORTED:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 			  ecpg_gettext("unsupported type \"%s\" on line %d"), str, line);
 			break;
 
 		case ECPG_TOO_MANY_ARGUMENTS:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("too many arguments on line %d"), line);
 			break;
 
 		case ECPG_TOO_FEW_ARGUMENTS:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("too few arguments on line %d"), line);
 			break;
 
 		case ECPG_INT_FORMAT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("invalid input syntax for type int: \"%s\", on line %d"), str, line);
 			break;
 
 		case ECPG_UINT_FORMAT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("invalid input syntax for type unsigned int: \"%s\", on line %d"), str, line);
 			break;
 
 		case ECPG_FLOAT_FORMAT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("invalid input syntax for floating-point type: \"%s\", on line %d"), str, line);
 			break;
 
 		case ECPG_CONVERT_BOOL:
 			if (str)
 				snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-				/*
-				 * translator: this string will be truncated at 149 characters
-				 * expanded.
-				 */
+				/*------
+				   translator: this string will be truncated at 149 characters expanded.  */
 						 ecpg_gettext("invalid syntax for type boolean: \"%s\", on line %d"), str, line);
 			else
 				snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-				/*
-				 * translator: this string will be truncated at 149 characters
-				 * expanded.
-				 */
+				/*------
+				   translator: this string will be truncated at 149 characters expanded.  */
 						 ecpg_gettext("could not convert boolean value: size mismatch, on line %d"), line);
 			break;
 
 		case ECPG_EMPTY:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("empty query on line %d"), line);
 			break;
 
 		case ECPG_MISSING_INDICATOR:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 			  ecpg_gettext("null value without indicator on line %d"), line);
 			break;
 
 		case ECPG_NO_ARRAY:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("variable does not have an array type on line %d"), line);
 			break;
 
 		case ECPG_DATA_NOT_ARRAY:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("data read from server is not an array on line %d"), line);
 			break;
 
 		case ECPG_ARRAY_INSERT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("inserting an array of variables is not supported on line %d"), line);
 			break;
 
 		case ECPG_NO_CONN:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("connection \"%s\" does not exist on line %d"), str, line);
 			break;
 
 		case ECPG_NOT_CONN:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("not connected to connection \"%s\" on line %d"), str, line);
 			break;
 
 		case ECPG_INVALID_STMT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("invalid statement name \"%s\" on line %d"), str, line);
 			break;
 
 		case ECPG_UNKNOWN_DESCRIPTOR:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("descriptor \"%s\" not found on line %d"), str, line);
 			break;
 
 		case ECPG_INVALID_DESCRIPTOR_INDEX:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 			 ecpg_gettext("descriptor index out of range on line %d"), line);
 			break;
 
 		case ECPG_UNKNOWN_DESCRIPTOR_ITEM:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("unrecognized descriptor item \"%s\" on line %d"), str, line);
 			break;
 
 		case ECPG_VAR_NOT_NUMERIC:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("variable does not have a numeric type on line %d"), line);
 			break;
 
 		case ECPG_VAR_NOT_CHAR:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("variable does not have a character type on line %d"), line);
 			break;
 
 		case ECPG_TRANS:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 			ecpg_gettext("error in transaction processing on line %d"), line);
 			break;
 
 		case ECPG_CONNECT:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("could not connect to database \"%s\" on line %d"), str, line);
 			break;
 
 		default:
 			snprintf(sqlca->sqlerrm.sqlerrmc, sizeof(sqlca->sqlerrm.sqlerrmc),
-
-			/*
-			 * translator: this string will be truncated at 149 characters
-			 * expanded.
-			 */
+			/*------
+			   translator: this string will be truncated at 149 characters expanded.  */
 					 ecpg_gettext("SQL error %d on line %d"), code, line);
 			break;
 	}
@@ -349,8 +271,8 @@ ecpg_raise_backend(int line, PGresult *result, PGconn *conn, int compat)
 		sqlca->sqlcode = ECPG_PGSQL;
 
 	/* %.*s is safe here as long as sqlstate is all-ASCII */
-	ecpg_log("raising sqlstate %.*s (sqlcode %d): %s\n",
-			 sizeof(sqlca->sqlstate), sqlca->sqlstate, sqlca->sqlcode, sqlca->sqlerrm.sqlerrmc);
+	ecpg_log("raising sqlstate %.*s (sqlcode %ld): %s\n",
+			 (int) sizeof(sqlca->sqlstate), sqlca->sqlstate, sqlca->sqlcode, sqlca->sqlerrm.sqlerrmc);
 
 	/* free all memory we have allocated for the user */
 	ECPGfree_auto_mem();
