@@ -452,7 +452,7 @@ dump_variables(struct arguments * list, int mode)
 	dump_variables(list->next, mode);
 
 	/* Then the current element and its indicator */
-	ECPGdump_a_type(yyout, list->variable->name, list->variable->type, list->variable->brace_level,
+	ECPGdump_a_type(base_yyout, list->variable->name, list->variable->type, list->variable->brace_level,
 					list->indicator->name, list->indicator->type, list->indicator->brace_level,
 					NULL, NULL, str_zero, NULL, NULL);
 
