@@ -3,7 +3,7 @@
  * fe-lobj.c
  *	  Front-end large object interface
  *
- * Portions Copyright (c) 1996-2014, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2015, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -290,7 +290,7 @@ lo_read(PGconn *conn, int fd, char *buf, size_t len)
 	argv[1].u.integer = (int) len;
 
 	res = PQfn(conn, conn->lobjfuncs->fn_lo_read,
-			   (int *) buf, &result_len, 0, argv, 2);
+			   (void *) buf, &result_len, 0, argv, 2);
 	if (PQresultStatus(res) == PGRES_COMMAND_OK)
 	{
 		PQclear(res);
@@ -441,8 +441,8 @@ lo_lseek64(PGconn *conn, int fd, pg_int64 offset, int whence)
 	argv[2].u.integer = whence;
 
 	res = PQfn(conn, conn->lobjfuncs->fn_lo_lseek64,
-			   (int *) &retval, &result_len, 0, argv, 3);
-	if (PQresultStatus(res) == PGRES_COMMAND_OK)
+			   (void *) &retval, &result_len, 0, argv, 3);
+	if (PQresultStatus(res) == PGRES_COMMAND_OK && result_len == 8)
 	{
 		PQclear(res);
 		return lo_ntoh64(retval);
@@ -607,8 +607,8 @@ lo_tell64(PGconn *conn, int fd)
 	argv[0].u.integer = fd;
 
 	res = PQfn(conn, conn->lobjfuncs->fn_lo_tell64,
-			   (int *) &retval, &result_len, 0, argv, 1);
-	if (PQresultStatus(res) == PGRES_COMMAND_OK)
+			   (void *) &retval, &result_len, 0, argv, 1);
+	if (PQresultStatus(res) == PGRES_COMMAND_OK && result_len == 8)
 	{
 		PQclear(res);
 		return lo_ntoh64(retval);
