@@ -22,6 +22,7 @@
 #include <nodes/makefuncs.h>
 #include <nodes/makefuncs.h>
 #include <parser/parse_clause.h>
+#include <parser/parse_collate.h>
 #include <parser/parse_node.h>
 #include <parser/parse_relation.h>
 #include <rewrite/rowsecurity.h>
@@ -107,6 +108,8 @@ test_rls_hooks_permissive(CmdType cmdtype, Relation relation)
 	policy->qual = (Expr *) transformWhereClause(qual_pstate, copyObject(e),
 												 EXPR_KIND_POLICY,
 												 "POLICY");
+	/* Fix up collation information */
+	assign_expr_collations(qual_pstate, (Node *) policy->qual);
 
 	policy->with_check_qual = copyObject(policy->qual);
 	policy->hassublinks = false;
@@ -165,6 +168,8 @@ test_rls_hooks_restrictive(CmdType cmdtype, Relation relation)
 	policy->qual = (Expr *) transformWhereClause(qual_pstate, copyObject(e),
 												 EXPR_KIND_POLICY,
 												 "POLICY");
+	/* Fix up collation information */
+	assign_expr_collations(qual_pstate, (Node *) policy->qual);
 
 	policy->with_check_qual = copyObject(policy->qual);
 	policy->hassublinks = false;
